'''
Author: Dr. Mohamed A. Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.
'''

from __future__ import division
import warnings
import numpy as np
from sklearn.cross_decomposition.pls_ import PLSRegression as pls
from smt.surrogate_models.krg_based import KrgBased
from smt.utils.kriging_utils import  componentwise_distance_PLS, componentwise_distance

"""
The KPLS class.
"""

class KPLSK(KrgBased):

    """
    - KPLSK
    """
    def _initialize(self):
        super(KPLSK, self)._initialize()
        declare = self.options.declare
        declare('n_comp', 1, types=int, desc='Number of principal components')
        declare('theta0', [1e-2], types=(list, np.ndarray), desc='Initial hyperparameters')
        self.name = 'KPLSK'

    def _compute_pls(self,X,y):
        _pls = pls(self.options['n_comp'])
        self.coeff_pls = _pls.fit(X.copy(),y.copy()).x_rotations_

        return X,y

    def _componentwise_distance(self,dx,opt=0):
        if opt == 0:
            # Kriging step
            d = componentwise_distance(dx,self.options['corr'].__name__,self.nx)
        else:
            # KPLS step
            d = componentwise_distance_PLS(dx,self.options['corr'].__name__,
                                                self.options['n_comp'],self.coeff_pls)
        return d



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.
"""

from __future__ import division
from smt.utils.options_dictionary import OptionsDictionary

from smt.surrogate_models import LS, QP, KPLS, KRG, KPLSK, GEKPLS
try:
    from smt.surrogate_models import IDW, RBF, RMTC, RMTB
    compiled_available = True
except:
    compiled_available = False

class Extensions(object):

    if compiled_available:
        _surrogate_type = {
            'KRG': KRG,'LS': LS,'QP': QP,'KPLS':KPLS,'KPLSK':KPLSK,'GEKPLS':GEKPLS,
            'RBF':RBF,'RMTC':RMTC,'RMTB':RMTB,'IDW':IDW}
    else:
        _surrogate_type = {
            'KRG': KRG,'LS': LS,'QP': QP,'KPLS':KPLS,'KPLSK':KPLSK,'GEKPLS':GEKPLS}

    def __init__(self, **kwargs):
        """
        Constructor where values of options can be passed in.

        For the list of options, see the documentation for the surrogate model being used.

        Parameters
        ----------
        **kwargs : named arguments
            Set of options that can be optionally set; each option must have been declared.

        Examples
        --------
        >>> from smt.extensions import VFM
        >>> extension = VFM(type_bridge = 'Additive', name_model_LF = QP, name_model_bridge =
                           LS, X_LF = xLF, y_LF = yLF, X_HF = xHF, y_HF = yHF, options_LF =
                           dictOptionLFModel, options_bridge = dictOptionBridgeModel)
        """
        self.options = OptionsDictionary()

        self._initialize()
        self.options.update(kwargs)

    def _initialize(self):
        """
        Implemented by the application to declare options and declare what they support (optional).

        Examples
        --------
        self.options.declare('option_name', default_value, types=(bool, int), desc='description')
        """
        pass

    def apply_method(self):
        """
        Run the complete algorithm of the SMT application; e.g.: VFM, ME, EGO...

        """
        self._apply()

    def analyse_results(self, **kwargs):
        """
        Get the final results; e.g., for VFM, two possible analysis are available:
        - kwargs = {x = x, operation = 'predict_values'}
        - kwargs = {x = x, operation = 'predict_derivatives, kx = i}

        """
        return self._analyse_results(**kwargs)



'''
Author: Dr. John T. Hwang <hwangjt@umich.edu>
        
This package is distributed under New BSD license.
'''

import numpy as np

def check_2d_array(array, name):
    if not isinstance(array, np.ndarray):
        raise ValueError('{} must be a NumPy array'.format(name))

    array = np.atleast_2d(array.T).T

    if len(array.shape) != 2:
        raise ValueError('{} must have a rank of 1 or 2'.format(name))

    return array

def check_support(sm, name, fail=False):
    if not sm.supports[name] or fail:
        class_name = sm.__class__.__name__
        raise NotImplementedError('{} does not support {}'.format(class_name, name))

def check_nx(nx, x):
    if x.shape[1] != nx:
        if nx == 1:
            raise ValueError('x should have shape [:, 1] or [:]')
        else:
            raise ValueError('x should have shape [:, {}] and not {}'.format(nx, x.shape))



'''
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.
'''

from __future__ import division

import numpy as np
import scipy.sparse
from six.moves import range
from numbers import Integral

from smt.utils.linear_solvers import get_solver, LinearSolver, VALID_SOLVERS
from smt.utils.line_search import get_line_search_class, LineSearch, VALID_LINE_SEARCHES
from smt.utils.caching import cached_operation
from smt.surrogate_models.surrogate_model import SurrogateModel


class RMTS(SurrogateModel):
    """
    Regularized Minimal-energy Tensor-product Spline interpolant base class for RMTC and RMTB.
    """

    def _initialize(self):
        super(RMTS, self)._initialize()
        declare = self.options.declare
        supports = self.supports

        declare('xlimits', types=np.ndarray,
                desc='Lower/upper bounds in each dimension - ndarray [nx, 2]')
        declare('smoothness', 1.0, types=(Integral, float, tuple, list, np.ndarray),
                desc='Smoothness parameter in each dimension - length nx. None implies uniform')
        declare('regularization_weight', 1e-14, types=(Integral, float),
                desc='Weight of the term penalizing the norm of the spline coefficients.' +
                    ' This is useful as an alternative to energy minimization ' +
                    ' when energy minimization makes the training time too long.')
        declare('energy_weight', 1e-4, types=(Integral, float),
                desc='The weight of the energy minimization terms')
        declare('extrapolate', False, types=bool,
                desc='Whether to perform linear extrapolation for external evaluation points')
        declare('min_energy', True, types=bool,
                desc='Whether to perform energy minimization')
        declare('approx_order', 4, types=Integral,
                desc='Exponent in the approximation term')
        declare('solver', 'krylov', values=VALID_SOLVERS, types=LinearSolver,
                desc='Linear solver')
        declare('derivative_solver', 'krylov', values=VALID_SOLVERS, types=LinearSolver,
                desc='Linear solver used for computing output derivatives (dy_dyt)')
        declare('grad_weight', 0.5, types=(Integral, float),
                desc='Weight on gradient training data')
        declare('solver_tolerance', 1e-12, types=(Integral, float),
                desc='Convergence tolerance for the nonlinear solver')
        declare('nonlinear_maxiter', 10, types=Integral,
                desc='Maximum number of nonlinear solver iterations')
        declare('line_search', 'backtracking', values=VALID_LINE_SEARCHES, types=LineSearch,
                desc='Line search algorithm')
        declare('save_energy_terms', False, types=bool,
                desc='Whether to cache energy terms in the data_dir directory')
        declare('data_dir', None, values=(None,), types=str,
                desc='Directory for loading / saving cached data; None means do not save or load')
        declare('max_print_depth', 5, types=Integral,
                desc='Maximum depth (level of nesting) to print operation descriptions and times')

        supports['training_derivatives'] = True
        supports['derivatives'] = True
        supports['output_derivatives'] = True

    def _setup_hessian(self):
        diag = np.ones(self.num['dof'])
        arange = np.arange(self.num['dof'])
        full_hess = scipy.sparse.csc_matrix((diag, (arange, arange)))
        return full_hess

    def _compute_jac(self, ix1, ix2, x):
        data, rows, cols = self._compute_jac_raw(ix1, ix2, x)
        n = x.shape[0]
        full_jac = scipy.sparse.csc_matrix((data, (rows, cols)), shape=(n, self.num['coeff']))
        if self.full_dof2coeff is not None:
            full_jac = full_jac * self.full_dof2coeff
        return full_jac

    def _compute_approx_terms(self):
        # This computes the approximation terms for the training points.
        # We loop over kx: 0 is for values and kx>0 represents.
        # the 1-based index of the derivative given by the training point data.
        num = self.num
        xlimits = self.options['xlimits']

        full_jac_dict = {}
        for kx in self.training_points[None]:
            xt, yt = self.training_points[None][kx]

            xmin = np.min(xt, axis=0)
            xmax = np.max(xt, axis=0)
            assert np.all(xlimits[:, 0] <= xmin), 'Training points below min for %s' % kx
            assert np.all(xlimits[:, 1] >= xmax), 'Training points above max for %s' % kx

            if kx == 0:
                c = 1.0
            else:
                self.options['grad_weight'] / xlimits.shape[0]

            full_jac = self._compute_jac(kx, 0, xt)
            full_jac_dict[kx] = (full_jac, full_jac.T.tocsc(), c)

        return full_jac_dict

    def _compute_energy_terms(self):
        # This computes the energy terms that are to be minimized.
        # The quadrature points are the centroids of the multi-dimensional elements.
        num = self.num
        xlimits = self.options['xlimits']

        inputs = {}
        inputs['nx'] = xlimits.shape[0]
        inputs['elem_list'] = num['elem_list']
        if self.__class__.__name__ == 'RMTB':
            inputs['num_ctrl_list'] = num['ctrl_list']
            inputs['order_list'] = num['order_list']

        if self.options['save_energy_terms']:
            cache_dir = self.options['data_dir']
        else:
            cache_dir = None
        with cached_operation(inputs, cache_dir) as outputs:
            if outputs:
                sq_mtx = outputs['sq_mtx']
            else:
                n = np.prod(2 * num['elem_list'])
                x = np.empty(n * num['x'])
                self.rmtsc.compute_quadrature_points(
                    n, np.array(2 * num['elem_list'], dtype=np.int32), x)
                x = x.reshape((n, num['x']))

                sq_mtx = [None] * num['x']
                for kx in range(num['x']):
                    mtx = self._compute_jac(kx+1, kx+1, x)
                    sq_mtx[kx] = mtx.T.tocsc() * mtx * (xlimits[kx, 1] - xlimits[kx, 0]) ** 4

                outputs['sq_mtx'] = sq_mtx

        elem_vol = np.prod((xlimits[:, 1] - xlimits[:, 0]) / (2 * num['elem_list']))
        total_vol = np.prod(xlimits[:, 1] - xlimits[:, 0])

        full_hess = scipy.sparse.csc_matrix((num['dof'], num['dof']))
        for kx in range(num['x']):
            full_hess += sq_mtx[kx] * (elem_vol / total_vol * self.options['smoothness'][kx] \
                / (xlimits[kx, 1] - xlimits[kx, 0]) ** 4)

        return full_hess

    def _opt_func(self, sol, p, yt_dict):
        full_hess = self.full_hess
        full_jac_dict = self.full_jac_dict

        func = 0.5 * np.dot(sol, full_hess * sol)
        for kx in self.training_points[None]:
            full_jac, full_jac_T, c = full_jac_dict[kx]
            yt = yt_dict[kx]
            func += 0.5 * c * np.sum((full_jac * sol - yt) ** p)

        return func

    def _opt_grad(self, sol, p, yt_dict):
        full_hess = self.full_hess
        full_jac_dict = self.full_jac_dict

        grad = full_hess * sol
        for kx in self.training_points[None]:
            full_jac, full_jac_T, c = full_jac_dict[kx]
            yt = yt_dict[kx]
            grad += 0.5 * c * full_jac_T * p * (full_jac * sol - yt) ** (p - 1)

        return grad

    def _opt_dgrad_dyt(self, sol, p, yt_dict, kx):
        full_hess = self.full_hess
        full_jac_dict = self.full_jac_dict

        full_jac, full_jac_T, c = full_jac_dict[kx]
        yt = yt_dict[kx]

        diag_vec = p * (p - 1) * (full_jac * sol - yt) ** (p - 2)
        diag_mtx = scipy.sparse.diags(diag_vec, format='csc')
        mtx = 0.5 * c * full_jac_T.dot(diag_mtx)

        return -mtx.todense()

    def _opt_hess(self, sol, p, yt_dict):
        full_hess = self.full_hess
        full_jac_dict = self.full_jac_dict

        hess = scipy.sparse.csc_matrix(full_hess)
        for kx in self.training_points[None]:
            full_jac, full_jac_T, c = full_jac_dict[kx]
            yt = yt_dict[kx]

            diag_vec = p * (p - 1) * (full_jac * sol - yt) ** (p - 2)
            diag_mtx = scipy.sparse.diags(diag_vec, format='csc')
            hess += 0.5 * c * full_jac_T * diag_mtx * full_jac

        return hess

    def _opt_norm(self, sol, p, yt_dict):
        full_hess = self.full_hess
        full_jac_dict = self.full_jac_dict

        grad = self._opt_grad(sol, p, yt_dict)
        return np.linalg.norm(grad)

    def _get_yt_dict(self, ind_y):
        yt_dict = {}
        for kx in self.training_points[None]:
            xt, yt = self.training_points[None][kx]
            yt_dict[kx] = yt[:, ind_y]
        return yt_dict

    def _run_newton_solver(self, sol):
        num = self.num
        options = self.options

        solver = get_solver(options['solver'])
        ls_class = get_line_search_class(options['line_search'])

        total_size = int(num['dof'])
        rhs = np.zeros((total_size, num['y']))
        d_sol = np.zeros((total_size, num['y']))

        p = self.options['approx_order']
        for ind_y in range(rhs.shape[1]):

            with self.printer._timed_context('Solving for output %i' % ind_y):

                yt_dict = self._get_yt_dict(ind_y)

                norm = self._opt_norm(sol[:, ind_y], p, yt_dict)
                fval = self._opt_func(sol[:, ind_y], p, yt_dict)
                self.printer(
                    'Iteration (num., iy, grad. norm, func.) : %3i %3i %15.9e %15.9e'
                    % (0, ind_y, norm, fval))

                iter_count = 0
                while iter_count < options['nonlinear_maxiter'] and norm > options['solver_tolerance']:
                    with self.printer._timed_context():
                        with self.printer._timed_context('Assembling linear system'):
                            mtx = self._opt_hess(sol[:, ind_y], p, yt_dict)
                            rhs[:, ind_y] = -self._opt_grad(sol[:, ind_y], p, yt_dict)

                        with self.printer._timed_context('Initializing linear solver'):
                            solver._setup(mtx, self.printer)

                        with self.printer._timed_context('Solving linear system'):
                            solver._solve(rhs[:, ind_y], d_sol[:, ind_y], ind_y=ind_y)

                        func = lambda x: self._opt_func(x, p, yt_dict)
                        grad = lambda x: self._opt_grad(x, p, yt_dict)

                        # sol[:, ind_y] += d_sol[:, ind_y]

                        ls = ls_class(sol[:, ind_y], d_sol[:, ind_y], func, grad)
                        with self.printer._timed_context('Performing line search'):
                            sol[:, ind_y] = ls(1.0)

                    norm = self._opt_norm(sol[:, ind_y], p, yt_dict)
                    fval = self._opt_func(sol[:, ind_y], p, yt_dict)
                    self.printer(
                        'Iteration (num., iy, grad. norm, func.) : %3i %3i %15.9e %15.9e'
                        % (iter_count, ind_y, norm, fval))

                    self.mtx = mtx

                    iter_count += 1

    def _solve(self):
        num = self.num
        options = self.options

        solver = get_solver(options['solver'])
        ls_class = get_line_search_class(options['line_search'])

        total_size = int(num['dof'])
        rhs = np.zeros((total_size, num['y']))
        sol = np.zeros((total_size, num['y']))
        d_sol = np.zeros((total_size, num['y']))

        with self.printer._timed_context('Solving initial startup problem (n=%i)' % total_size):

            approx_order = options['approx_order']
            nonlinear_maxiter = options['nonlinear_maxiter']
            options['approx_order'] = 2
            options['nonlinear_maxiter'] = 1

            self._run_newton_solver(sol)

            options['approx_order'] = approx_order
            options['nonlinear_maxiter'] = nonlinear_maxiter

        with self.printer._timed_context('Solving nonlinear problem (n=%i)' % total_size):

            self._run_newton_solver(sol)

        return sol

    def _new_train(self):
        """
        Train the model
        """
        with self.printer._timed_context('Pre-computing matrices', 'assembly'):

            with self.printer._timed_context('Computing dof2coeff', 'dof2coeff'):
                self.full_dof2coeff = self._compute_dof2coeff()

            with self.printer._timed_context('Initializing Hessian', 'init_hess'):
                self.full_hess = self._setup_hessian() * self.options['regularization_weight']

            if self.options['min_energy']:
                with self.printer._timed_context('Computing energy terms', 'energy'):
                    self.full_hess += self._compute_energy_terms() * self.options['energy_weight']

            with self.printer._timed_context('Computing approximation terms', 'approx'):
                self.full_jac_dict = self._compute_approx_terms()

        with self.printer._timed_context('Solving for degrees of freedom', 'total_solution'):
            self.sol = self._solve()

        if self.full_dof2coeff is not None:
            self.sol_coeff = self.full_dof2coeff * self.sol
        else:
            self.sol_coeff = self.sol

    def _train(self):
        """
        Train the model
        """
        self._setup()

        tmp = self.rmtsc
        self.rmtsc = None

        inputs = {'self': self}
        with cached_operation(inputs, self.options['data_dir']) as outputs:
            self.rmtsc = tmp

            if outputs:
                self.sol_coeff = outputs['sol_coeff']
                self.sol = outputs['sol']
                self.mtx = outputs['mtx']
                self.full_dof2coeff = outputs['full_dof2coeff']
                self.full_hess = outputs['full_hess']
                self.full_jac_dict = outputs['full_jac_dict']
            else:
                self._new_train()
                outputs['sol_coeff'] = self.sol_coeff
                outputs['sol'] = self.sol
                outputs['mtx'] = self.mtx
                outputs['full_dof2coeff'] = self.full_dof2coeff
                outputs['full_hess'] = self.full_hess
                outputs['full_jac_dict'] = self.full_jac_dict

    def _predict_values(self, x):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        mtx = self._compute_prediction_mtx(x, 0)
        y = mtx.dot(self.sol_coeff)

        return y

    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray
            Derivative values.
        """
        mtx = self._compute_prediction_mtx(x, kx + 1)
        y = mtx.dot(self.sol_coeff)

        return y

    def _compute_prediction_mtx(self, x, kx):
        n = x.shape[0]

        num = self.num
        options = self.options

        data, rows, cols = self._compute_jac_raw(kx, 0, x)

        # In the explanation below, n is the number of dimensions, and
        # a_k and b_k are the lower and upper bounds for x_k.
        #
        # A C1 extrapolation can get very tricky, so we implement a simple C0
        # extrapolation. We basically linarly extrapolate from the nearest
        # domain point. For example, if n = 4 and x2 > b2 and x3 > b3:
        #    f(x1,x2,x3,x4) = f(x1,b2,b3,x4) + dfdx2 (x2-b2) + dfdx3 (x3-b3)
        #    where the derivatives are evaluated at x1,b2,b3,x4 (called b) and
        #    dfdx1|x = dfdx1|b + d2fdx1dx2|b (x2-b2) + d2fdx1dx3|b (x3-b3)
        #    dfdx2|x = dfdx2|b.
        # The dfdx2|x derivative is what it is because f and all derivatives
        # evaluated at x1,b2,b3,x4 are constant with respect to changes in x2.
        # On the other hand, the dfdx1|x derivative is what it is because
        # f and all derivatives evaluated at x1,b2,b3,x4 change with x1.
        # The extrapolation function is non-differentiable at boundaries:
        # i.e., where x_k = a_k or x_k = b_k for at least one k.
        if options['extrapolate']:

            # First we evaluate the vector pointing to each evaluation points
            # from the nearest point on the domain, in a matrix called dx.
            # If the ith evaluation point is not external, dx[i, :] = 0.
            dx = np.empty(n * num['support'] * num['x'])
            self.rmtsc.compute_ext_dist(n, num['support'], x.flatten(), dx)
            dx = dx.reshape((n * num['support'], num['x']))

            isexternal = np.array(np.array(dx, bool), float)

            for ix in range(num['x']):
                # Now we compute the first order term where we have a
                # derivative times (x_k - b_k) or (x_k - a_k).
                data_tmp, rows, cols = self._compute_jac_raw(kx, ix+1, x)
                data_tmp *= dx[:, ix]

                # If we are evaluating a derivative (with index kx),
                # we zero the first order terms for which dx_k = 0.
                if kx != 0:
                    data_tmp *= 1 - isexternal[:, kx-1]

                data += data_tmp

        mtx = scipy.sparse.csc_matrix((data, (rows, cols)), shape=(n, num['coeff']))
        return mtx

    def _predict_output_derivatives(self, x):
        # dy_dyt = dy_dw * (dR_dw)^{-1} * dR_dyt

        n = x.shape[0]
        nw = self.mtx.shape[0]
        nx = x.shape[1]
        ny = self.sol.shape[1]

        p = self.options['approx_order']

        dy_dw = self._compute_prediction_mtx(x, 0)
        if self.full_dof2coeff is not None:
            dy_dw = dy_dw * self.full_dof2coeff
        dy_dw = dy_dw.todense()

        dR_dw = self.mtx

        dy_dyt = {}
        for kx in self.training_points[None]:
            nt = self.training_points[None][kx][0].shape[0]

            dR_dyt = np.zeros((nw, nt, ny))
            for ind_y in range(ny):
                yt_dict = self._get_yt_dict(ind_y)
                dR_dyt[:, :, ind_y] = self._opt_dgrad_dyt(self.sol[:, ind_y], p, yt_dict, kx)

            solver = get_solver(self.options['derivative_solver'])
            solver._setup(dR_dw, self.printer)

            dw_dyt = np.zeros((nw, nt, ny))
            for ind_t in range(nt):
                for ind_y in range(ny):
                    solver._solve(
                        dR_dyt[:, ind_t, ind_y],
                        dw_dyt[:, ind_t, ind_y],
                        ind_y=ind_y)
                    dw_dyt[:, ind_t, ind_y] *= -1.

            if kx == 0:
                dy_dyt[None] = np.einsum('ij,jkl->ikl', dy_dw, dw_dyt)
            else:
                dy_dyt[kx-1] = np.einsum('ij,jkl->ikl', dy_dw, dw_dyt)

        return dy_dyt



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

N-dimensional cantilever beam problem.
"""
import numpy as np
from six.moves import range

from smt.utils.options_dictionary import OptionsDictionary
from smt.problems.problem import Problem
from smt.problems.reduced_problem import ReducedProblem
from smt.problems.cantilever_beam import CantileverBeam


class NdimCantileverBeam(Problem):

    def __init__(self, ndim=1, w=0.2):
        self.problem = ReducedProblem(CantileverBeam(ndim=3*ndim), np.arange(1, 3*ndim, 3), w=w)

        self.options = OptionsDictionary()
        self.options.declare('ndim', ndim, types=int)
        self.options.declare('return_complex', False, types=bool)
        self.options.declare('name', 'NdimCantileverBeam', types=str)

        self.xlimits = self.problem.xlimits

    def _evaluate(self, x, kx):
        return self.problem._evaluate(x, kx)



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
        Dr. Nathalie.bartoli      <nathalie@onera.fr>

This package is distributed under New BSD license.

TO DO:
- define outputs['sol'] = self.sol
"""

from __future__ import division

import numpy as np

from sklearn import linear_model
from smt.surrogate_models.surrogate_model import SurrogateModel
from smt.utils.caching import cached_operation


class LS(SurrogateModel):

    """
    Least square model.
    This model uses the linear_model.LinearRegression class from scikit-learn.
    Default-parameters from scikit-learn are used herein.
    """

    def _initialize(self):
        super(LS, self)._initialize()
        declare = self.options.declare
        supports = self.supports

        declare('data_dir', values=None, types=str,
                desc='Directory for loading / saving cached data; None means do not save or load')

        self.name = 'LS'
        supports['derivatives'] = True

    ############################################################################
    # Model functions
    ############################################################################

    def _new_train(self):
        """
        Train the model
        """
        pts = self.training_points

        if 0 in pts[None]:
            x = pts[None][0][0]
            y = pts[None][0][1]

        self.mod = linear_model.LinearRegression()
        self.mod.fit(x,y)

    def _train(self):
        """
        Train the model
        """
        inputs = {'self': self}
        with cached_operation(inputs, self.options['data_dir']) as outputs:
            if outputs:
                self.sol = outputs['sol']
            else:
                self._new_train()
                #outputs['sol'] = self.sol

    def _predict_values(self,x):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        y = self.mod.predict(x)
        return y

    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray
            Derivative values.
        """

        # Initialization
        n_eval, n_features_x = x.shape
        y = np.ones((n_eval,1)) * self.mod.coef_[0,kx]

        return y



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.

Variable-fidelity modeling: two types of bridge functions are available; i.e.,
additive and multiplicative
"""

from __future__ import division
import numpy as np

from smt.utils.options_dictionary import OptionsDictionary
from types import FunctionType
from smt.extensions.extensions import Extensions

class VFM(Extensions):

    def _initialize(self):
        super(VFM, self)._initialize()
        declare = self.options.declare

        declare('name_model_LF',types=object,values=('KRG','LS','QP','KPLS',
                'KPLSK','GEKPLS','RBF','RMTC','RMTB','IDW'), desc=
                'Name of the low-fidelity model')
        declare('options_LF',{},types=dict,desc='Options for the low-fidelity model')
        declare('name_model_bridge', types=object, values=('KRG','LS','QP','KPLS',
                'KPLSK','GEKPLS','RBF','RMTC','RMTB','IDW'), desc=
                'Name of the bridge model')
        declare('options_bridge', {},types=dict,desc='Options for the bridge model')
        declare('type_bridge', 'Additive',types=str,values=('Additive', 'Multiplicative'), desc=
                'Bridge function type')
        declare('X_LF', None,types=np.ndarray,desc='Low-fidelity inputs')
        declare('y_LF', None,types=np.ndarray,desc='Low-fidelity output')
        declare('X_HF', None,types=np.ndarray,desc='High-fidelity inputs')
        declare('y_HF', None,types=np.ndarray,desc='High-fidelity output')
        declare('dy_LF', None,types=np.ndarray,desc='Low-fidelity derivatives')
        declare('dy_HF', None,types=np.ndarray,desc='High-fidelity derivatives')

    def _apply(self):
        """
        Algorithm of the VFM method
        """

        # For seek of readability
        if self.options['X_LF'] is not None and self.options['y_LF'] is not None and \
           self.options['X_HF'] is not None and self.options['y_HF'] is not None:
            X_LF = self.options['X_LF']
            y_LF = self.options['y_LF']
            X_HF = self.options['X_HF']
            y_HF = self.options['y_HF']
        else:
            raise ValueError('Check X_LF, y_LF, X_HF, and y_FH')

        if self.options['dy_LF'] is not None:
            dy_LF = self.options['dy_LF']
        if self.options['dy_HF'] is not None:
            dy_HF = self.options['dy_HF']

        # Check parameters
        self._check_param()

        #Train the low fidelity model
        self.LF_deriv = self.options['options_LF']['deriv']
        del self.options['options_LF']['deriv']
        sm_LF = self.options['name_model_LF'](**self.options['options_LF'])
        sm_LF.set_training_values(X_LF,y_LF)
        if self.LF_deriv:
            for i in range(sm_LF.nx):
                sm_LF.set_training_derivatives(X_LF,dy_LF[:,i],i)

        sm_LF.train()

        # compute the bridge data
        if self.options['type_bridge'] == 'Multiplicative':
            y_bridge = y_HF/sm_LF.predict_values(X_HF)
            self.B_deriv = self.options['options_bridge']['deriv']
            del self.options['options_bridge']['deriv']
            if self.B_deriv:
                dy_bridge = np.zeros(dy_HF.shape)
                for i in range(sm_LF.nx):
                    dy_bridge[:,i] = (((dy_HF[:,i].reshape((y_HF.shape[0],1))*
                        sm_LF.predict_values(X_HF))- (y_HF*sm_LF.predict_derivatives(X_HF,
                        kx=i)))/sm_LF.predict_values(X_HF)**2).reshape(y_HF.shape[0])
        elif self.options['type_bridge'] == 'Additive':
            y_bridge = y_HF-sm_LF.predict_values(X_HF)
            self.B_deriv = self.options['options_bridge']['deriv']
            del self.options['options_bridge']['deriv']
            if self.B_deriv:
                dy_bridge = np.zeros(dy_HF.shape)
                for i in range(sm_LF.nx):
                    dy_bridge[:,i] = (dy_HF[:,i].reshape((y_HF.shape[0],1))- \
                                      sm_LF.predict_derivatives(X_HF,kx=i)).reshape((y_HF.shape[0]))
        else:
            raise ValueError('Only Additive and Multiplicative bridges are available')

        # Construct of the bridge function
        sm_bridge = self.options['name_model_bridge'](**self.options['options_bridge'])
        sm_bridge.set_training_values(X_HF,y_bridge)
        if self.B_deriv and self.options['dy_HF'] is not None:
            for i in range(sm_bridge.nx):
                sm_bridge.set_training_derivatives(X_HF,dy_bridge[:,i],i)
        sm_bridge.train()

        # Construct the final model
        sm_HF = {}
        if self.options['type_bridge'] == 'Multiplicative':
            sm_HF['predict_values'] = lambda x: sm_bridge.predict_values(x)*sm_LF.predict_values(x)
            if sm_bridge.supports['derivatives'] and sm_LF.supports['derivatives']:
                sm_HF['predict_derivatives'] = []
                for i in range(sm_LF.nx):
                    sm_HF['predict_derivatives'].append(
                        lambda x,i=i: sm_bridge.predict_derivatives(x,i)*
                        sm_LF.predict_values(x) + sm_bridge.predict_values(x)*
                        sm_LF.predict_derivatives(x,i))

        else:
            sm_HF['predict_values'] = lambda x: sm_bridge.predict_values(x)+sm_LF.predict_values(x)
            if sm_bridge.supports['derivatives'] and sm_LF.supports['derivatives']:
                sm_HF['predict_derivatives'] = []
                for i in range(sm_LF.nx):
                    sm_HF['predict_derivatives'].append(
                        lambda x,i=i: sm_bridge.predict_derivatives(x,i) +
                        sm_LF.predict_derivatives(x,i))

        self.sm_HF = sm_HF

    def _analyse_results(self,x,operation = 'predict_values',kx=None):
        """
        Analyse the results requested by the user

        Parameters
        ----------
        x: np.ndarray[n, nx] or np.ndarray[n]
           Input values for the prediction result analysis.

        operation: str
           Type of the analysis. Two values are available: 'predict_values' or 'predict_derivatives'

        kx : int
           The 0-based index of the input variable with respect to which derivatives are desired.

        return
        ------
        y: np.ndarray
            Output values at the prediction value/derivative points.

        """

        if operation == 'predict_values':
            y = self.sm_HF['predict_values'](x)
        elif operation == 'predict_derivatives':
            if kx is None:
                y = np.zeros(x.shape)
                for i in range(x.shape[1]):
                    y[:,i] = self.sm_HF['predict_derivatives'][i](x).reshape((x.shape[0]))
            else:
                y = self.sm_HF['predict_derivatives'][kx](x).reshape((x.shape[0],1))

        return y

    def _check_param(self):

        """
        This function check some parameters of the model.
        """
        # Check surrogates
        if not callable(self.options['name_model_LF']):
            if self.options['name_model_LF'] in self._surrogate_type:
                self.options['name_model_LF'] = self._surrogate_type[
                    self.options['name_model_LF']]
            else:
                raise ValueError("The LF surrogate should be one of %s, "
                                 "%s was given." % (self._surrogate_type.keys(),
                                self.options['name_model_LF']))

        if not callable(self.options['name_model_bridge']):
            if self.options['name_model_bridge'] in self._surrogate_type:
                self.options['name_model_bridge'] = self._surrogate_type[
                    self.options['name_model_bridge']]
            else:
                raise ValueError("The bridge surrogate should be one of %s, "
                                 "%s was given." % (self._surrogate_type.keys(),
                                self.options['name_model_bridge']))

        # Initialize the parameter deriv
        if ('deriv' not in self.options['options_LF'].keys()):
            self.options['options_LF']['deriv'] = False
        if ('deriv' not in self.options['options_bridge'].keys()):
            self.options['options_bridge']['deriv'] = False
